/*
  postgresql.c -- Emacs Lisp binding to libpq.so
  Copyright (C) 2000 Electrotechnical Laboratory, JAPAN.
  Licensed to the Free Software Foundation.

Implementation notes:
0. Supported PostgreSQL versions
   This code was developed against libpq-6.5.3 and libpq-7.0-beta1.  Earlier
   versions may work.  V7 support is more complete than V6.5 support.
1. Mule
   This code won't work with non-ASCII databases and libpq-6.5 (or earlier).
   It will work with libpq-7.0 or later, but depends upon being able to set
   the client encoding to Mule Internal format.
2. Asynchronous Operation
   Starting with libpq-7.0, an asynchronous interface is offered.  This
   binding supports the asynchronous calls to a limited extent.  Since the
   XEmacs 21.2 core does not support a sensible interface to add managed but
   unreadable (by XEmacs) file descriptors to the main select code, polling
   is required to drive the asynchronous calls.  XtAppAddInput would work
   fine, but we want to be able to use the database when running strictly in
   tty mode.
3. Completeness
   Various calls have been deliberately not exported to Lisp.  The
   unexported calls are either left-over backwards compatibility code that
   aren't needed, calls that cannot be implemented sensibly, or calls that
   cannot be implemented safely.  A list of all global functions in libpq
   but not exported to Lisp is below.
4. Policy
   This interface tries very hard to not set any policy towards how database
   code in Emacs Lisp will be written.
5. Documentation
   For full lisp programming documentation, see the XEmacs Lisp Reference
   Manual.  For PostgreSQL documentation, see the PostgreSQL distribution.

TODO (in rough order of priority):
0. Fully implement signaling on error conditions.
1. Robust MULE-ization.
2. Asynchronous notifies need to be implemented to the extent they can be.
3. The large object interface needs work with Emacs buffers in addition
   to files.
4. The copy in/out functions either need to be implemented or moved to
   the unsupported list.
5. Do the setenv* functions need to be implemented?
*/

/*
  Unimplemented functions: [TODO]
  PQsetNoticeProcessor
  PQgetline (copy in/out)
  PQputline (copy in/out)
  PQgetlineAsync (copy in/out Asynch.)
  PQputnbytes (copy in/out Asynch.)
  PQendcopy (copy in/out)
  PQmakeEmptyPGresult [*]

  Unsupported functions:
  PQsetdbLogin -- This function is deprecated, has a subset of the
   functionality of PQconnectdb, and is better done in Lisp.
  PQsetdb -- Same as for PQsetdbLogin
  PQsocket -- Abstraction error, file descriptors should not be leaked
   into Lisp code
  PQprint -- print to a file descriptor, deprecated, better done in Lisp
  PQdisplayTuples -- deprecated
  PQprintTuples -- really, really deprecated
  PQsetenvStart (Asynch. Queries)
  PQsetenvPoll (Asynch. Queries)
  PQsetenvHandle (Asynch. Queries)
  PQsetenv -- The setenv functions are called as a normal part of connecting
   to a database.  With the possible exception of PQsetenv, it doesn't seem
   useful to pass on this interface.
  PQmblen -- Returns the length in bytes of multibyte character encoded
   string.
  PQtrace -- controls debug print tracing to a tty.
  PQuntrace -- Ditto.  I don't see any way to do this sensibly.
  PQoidStatus -- deprecated and nearly identical to PQoidValue
  PQfn -- "Fast path" interface
  lo_open (large object) [*]
  lo_close (large object) [*]
  lo_read (large object) [*]
  lo_write (large object) [*]
  lo_lseek (large object) [*]
  lo_creat (large object) [*]
  lo_tell (large object) [*]
  lo_unlink (large object) [*]
*/

#include <config.h>
#include "lisp.h"
#include "sysdep.h"

#include <libpq-fe.h>
#include "postgresql.h"

static Lisp_Object VXPGHOST;
static Lisp_Object VXPGUSER;
static Lisp_Object VXPGOPTIONS;
static Lisp_Object VXPGPORT;
static Lisp_Object VXPGTTY; /* This needs to be blanked! */
static Lisp_Object VXPGDATABASE;
static Lisp_Object VXPGREALM;
#ifdef MULE
static Lisp_Object VXPGCLIENTENCODING;
#endif /* MULE */

/* Other variables:
   PGAUTHTYPE -- not used after PostgreSQL 6.5
   PGGEQO
   PGCOSTINDEX
   PGCOSTHEAP
   PGTZ
   PGDATESTYLE
*/
#ifndef HAVE_POSTGRESQLV7
static Lisp_Object VXPGAUTHTYPE;
#endif
static Lisp_Object VXPGGEQO, VXPGCOSTINDEX, VXPGCOSTHEAP, VXPGTZ, VXPGDATESTYLE;

static Lisp_Object Qpg_connection_ok, Qpg_connection_bad;
static Lisp_Object Qpg_connection_started, Qpg_connection_made;
static Lisp_Object Qpg_connection_awaiting_response, Qpg_connection_auth_ok;
static Lisp_Object Qpg_connection_setenv;

static Lisp_Object Qpqdb, Qpquser, Qpqpass, Qpqhost, Qpqport, Qpqtty;
static Lisp_Object Qpqoptions, Qpqstatus, Qpqerrormessage, Qpqbackendpid;

static Lisp_Object Qpgres_empty_query, Qpgres_command_ok, Qpgres_tuples_ok;
static Lisp_Object Qpgres_copy_out, Qpgres_copy_in, Qpgres_bad_response;
static Lisp_Object Qpgres_nonfatal_error, Qpgres_fatal_error;

static Lisp_Object Qpgres_polling_failed, Qpgres_polling_reading;
static Lisp_Object Qpgres_polling_writing, Qpgres_polling_ok;
static Lisp_Object Qpgres_polling_active;
/****/

/* PGconn is an opaque object and we need to be able to store them in
   Lisp code because libpq supports multiple connections.
*/
Lisp_Object Qpgconnp;

static Lisp_Object
make_pgconn (struct Lisp_PGconn *pgconn)
{
  Lisp_Object lisp_pgconn;
  XSETPGCONN (lisp_pgconn, pgconn);
  return lisp_pgconn;
}

static Lisp_Object
mark_pgconn (Lisp_Object obj, void (*markobj) (Lisp_Object))
{
  return Qnil;
}

static void
print_pgconn (Lisp_Object obj, Lisp_Object printcharfun, int escapeflag)
{
  char buf[256];
  PGconn *pgc;
  ConnStatusType cst;
  char *host="", *db="", *user="", *port="";

  pgc = (XPGCONN (obj))->pgconn;
  
  if ((cst = PQstatus (pgc)) == CONNECTION_OK)
    {
      if (!(host = PQhost (pgc)))
	host = "";
      port = PQport (pgc);
      db = PQdb (pgc);
      if (!(user = PQuser (pgc)))
	user = "";
      sprintf (buf, "#<PGconn %s:%s %s/%s>", /* evil! */
	      !strlen (host) ? "localhost" : host,
	      port,
	      user,
	      db);
    }
  else
    sprintf (buf, "#<PGconn BAD>");

  if (print_readably)
    error ("printing unreadable object %s", buf);
  else
    write_c_string (buf, printcharfun);
}

static struct Lisp_PGconn *
allocate_pgconn (void)
{
  struct Lisp_PGconn *pgconn = alloc_lcrecord_type (struct Lisp_PGconn,
						    lrecord_pgconn);
  pgconn->pgconn = (PGconn *)NULL;
  return pgconn;
}

static void
finalize_pgconn (void *header, int for_disksave)
{
  struct Lisp_PGconn *pgconn = (struct Lisp_PGconn *) header;

  if (for_disksave)
    signal_simple_error ("Can't dump an emacs containing PGconn objects",
                         make_pgconn (pgconn));

  if (pgconn->live)
    PQfinish (pgconn->pgconn);
}

DEFINE_LRECORD_IMPLEMENTATION ("pgconn", pgconn,
			       mark_pgconn, print_pgconn, finalize_pgconn,
			       NULL, NULL, struct Lisp_PGconn);

/****/

/* PGresult is an opaque object and we need to be able to store them in
   Lisp code.
*/
Lisp_Object Qpgresultp;

static Lisp_Object
make_pgresult (struct Lisp_PGresult *pgresult)
{
  Lisp_Object lisp_pgresult;
  XSETPGRESULT (lisp_pgresult, pgresult);
  return lisp_pgresult;
}

static Lisp_Object
mark_pgresult (Lisp_Object obj, void (*markobj) (Lisp_Object))
{
  return Qnil;
}

static void
print_pgresult (Lisp_Object obj, Lisp_Object printcharfun, int escapeflag)
{
  char buf[1024];
  PGresult *res;

  res = (XPGRESULT(obj))->pgresult;

  sprintf (buf, "#<PGresult %s - %s>",  /* evil! */
		 PQresStatus (PQresultStatus (res)),
		 PQcmdStatus (res));

  if (print_readably)
    error ("printing unreadable object %s", buf);
  else
    write_c_string (buf, printcharfun);
}

static struct Lisp_PGresult *
allocate_pgresult (void)
{
  struct Lisp_PGresult *pgresult = alloc_lcrecord_type (struct Lisp_PGresult,
							lrecord_pgresult);
  pgresult->pgresult = (PGresult *)NULL;
  return pgresult;
}

static void
finalize_pgresult (void *header, int for_disksave)
{
  struct Lisp_PGresult *pgresult = (struct Lisp_PGresult *) header;

  if (for_disksave)
    signal_simple_error ("Can't dump an emacs containing PGresult objects",
                         make_pgresult (pgresult));

  if (pgresult->live)
    PQclear (pgresult->pgresult);
}

DEFINE_LRECORD_IMPLEMENTATION ("pgresult", pgresult,
			       mark_pgresult, print_pgresult, finalize_pgresult,
			       NULL, NULL, struct Lisp_PGresult);


/* There are four ways (as of PostgreSQL v7) to connect to a database.
   Two of them, PQsetdb and PQsetdbLogin, are deprecated.  Both of those
   routines take a number of positional parameters and are better done in Lisp.
   Note that PQconnectStart does not exist prior to v7.
*/

DEFUN ("pq-conn-defaults", Fpq_conn_defaults, 0, 0, 0, /*
Return a connection default structure.
*/
       ())
{
  /* This function can GC */
  PQconninfoOption *pcio;
  Lisp_Object temp, temp1;
  int i;

  pcio = PQconndefaults();
  temp = list1 (Fcons (build_string (pcio[0].keyword),
		       Fcons (build_string (pcio[0].envvar),
			      Fcons (build_string (pcio[0].compiled),
				     Fcons (build_string (pcio[0].val),
					    Fcons (build_string (pcio[0].label),
						   Fcons (build_string (pcio[0].dispchar),
							  Fcons (make_int (pcio[0].dispsize), Qnil))))))));

  for (i = 1; pcio[i].keyword; i++)
    {
      temp1 = list1 (Fcons (build_string (pcio[i].keyword),
			    Fcons (build_string (pcio[i].envvar),
				   Fcons (build_string (pcio[i].compiled),
					  Fcons (build_string (pcio[i].val),
						 Fcons (build_string (pcio[i].label),
							Fcons (build_string (pcio[i].dispchar),
							       Fcons (make_int (pcio[i].dispsize), Qnil))))))));
      {
	Lisp_Object args[2];
	args[0] = temp;
	args[1] = temp1;
	/* Fappend GCPROs its arguments */
	temp = Fappend (2, args);
      }
    }

  return temp;
}

/* PQconnectdb Makes a new connection to a backend.
PGconn *PQconnectdb(const char *conninfo)
*/

DEFUN ("pq-connectdb", Fpq_connectdb, 1, 1, 0, /*
Make a new connection to a PostgreSQL backend.
*/
	(conninfo))
{
  PGconn *P;
  struct Lisp_PGconn *lisp_pgconn;

  CHECK_STRING (conninfo);

  P = PQconnectdb ((char *)XSTRING_DATA (conninfo));
  lisp_pgconn = allocate_pgconn();
  lisp_pgconn->pgconn = P;
  lisp_pgconn->live = 1;

  return make_pgconn (lisp_pgconn);
}

/* PQconnectStart Makes a new asynchronous connection to a backend.
PGconn *PQconnectStart(const char *conninfo)
*/

#ifdef HAVE_POSTGRESQLV7
DEFUN ("pq-connect-start", Fpq_connect_start, 1, 1, 0, /*
Make a new asynchronous connection to a PostgreSQL backend.
*/
	(conninfo))
{
  PGconn *P;
  struct Lisp_PGconn *lisp_pgconn;

  CHECK_STRING (conninfo);

  P = PQconnectStart ((char *)XSTRING_DATA (conninfo));

  if (P)
    {
      lisp_pgconn = allocate_pgconn();
      lisp_pgconn->pgconn = P;
      lisp_pgconn->live = 1;

      /* #### Here we need to set up an internal process handler */

      return make_pgconn (lisp_pgconn);
    }
  else
    return Qnil;
}

DEFUN ("pq-connect-poll", Fpq_connect_poll, 1, 1, 0, /*
Poll an asynchronous connection for completion
*/
	(conn))
{
  PGconn *P;
  PostgresPollingStatusType PS;

  CHECK_PGCONN (conn);

  P = (XPGCONN (conn))->pgconn;
  PS = PQconnectPoll (P);
  switch (PS)
    {
    case PGRES_POLLING_FAILED:
      return Qpgres_polling_failed;
    case PGRES_POLLING_OK:
      return Qpgres_polling_ok;
    case PGRES_POLLING_READING:
      return Qpgres_polling_reading;
    case PGRES_POLLING_WRITING:
      return Qpgres_polling_writing;
    case PGRES_POLLING_ACTIVE:
      return Qpgres_polling_active;
    default: abort (); /* they've added a new field, die */
    }

}

#ifdef MULE
DEFUN ("pq-client-encoding", Fpq_client_encoding, 1, 1, 0, /*
Return client coding system.
*/
       (conn))
{
  PGconn *P;

  CHECK_PGCONN (conn);
  P = (XPGCONN (conn))->pgconn;

  return make_int (PQclientEncoding (P));
}

DEFUN ("pq-set-client-encoding", Fpq_set_client_encoding, 2, 2, 0, /*
Set client coding system.
*/
       (conn, encoding))
{
  PGconn *P;

  CHECK_PGCONN (conn);
  CHECK_STRING (encoding);

  P = (XPGCONN (conn))->pgconn;

  return make_int (PQsetClientEncoding (P, XSTRING_DATA (encoding)));
}

#endif
#endif /* HAVE_POSTGRESQLV7 */

/* PQfinish Close the connection to the backend. Also frees memory
       used by the PGconn object.
void PQfinish(PGconn *conn)
*/
DEFUN ("pq-finish", Fpq_finish, 1, 1, 0, /*
Close the connection to the backend.
Note:  This entry point is provided for completeness, but its use is not
recommended.  The preferred way to delete connections is to let the Emacs
garbage collection mechanism do it.

DANGER:  This function isn't fully debugged and has been very prone to
crashes in testing.
*/
	(conn))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  PQfinish (pgconn);
  /* #### PQfinish deallocates the PGconn structure, so we now have a
     dangling pointer. */
  XSETPGCONN (conn, (PGconn *)NULL); /* will this work?   [no] */
  return Qnil;
}

DEFUN ("pq-clear", Fpq_clear, 1, 1, 0, /*
Forcibly erase a PGresult object.
*/
       (res))
{
  PGresult *pgr;

  CHECK_PGRESULT (res);
  pgr = (XPGRESULT (res))->pgresult;

  PQclear (pgr);
  XSETPGRESULT (res, (PGconn *)NULL); /* will this work? */
  return Qnil;
}

DEFUN ("pq-is-busy", Fpq_is_busy, 1, 1, 0, /*
Return t if PQgetResult would block waiting for input.
*/
	(conn))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  return PQisBusy (pgconn) ? Qt : Qnil;
}

DEFUN ("pq-consume-input", Fpq_consume_input, 1, 1, 0, /*
Consume any available input from the backend.
Returns nil if something bad happened.
*/
	(conn))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  return PQconsumeInput (pgconn) ? Qt : Qnil;
}

/* PQreset Reset the communication port with the backend.
void PQreset(PGconn *conn)
*/
DEFUN ("pq-reset", Fpq_reset, 1, 1, 0, /*
Reset the connection to the backend.
This function will close the connection to the backend and attempt to
reestablish a new connection to the same postmaster, using all the same
parameters previously used.  This may be useful for error recovery if a
working connection is lost.
*/
	(conn))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  PQreset (pgconn);

  return Qnil;
}

#ifdef HAVE_POSTGRESQLV7
DEFUN ("pq-reset-start", Fpq_reset_start, 1, 1, 0, /*
Reset connection to the backend asynchronously.
*/
       (conn))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  return PQresetStart (pgconn) ? Qt : Qnil;
}

DEFUN ("pq-reset-poll", Fpq_reset_poll, 1, 1, 0, /*
Poll an asynchronous reset for completion
*/
	(conn))
{
  PGconn *P;
  PostgresPollingStatusType PS;

  CHECK_PGCONN (conn);

  P = (XPGCONN (conn))->pgconn;
  PS = PQresetPoll (P);
  switch (PS)
    {
    case PGRES_POLLING_FAILED:
      return Qpgres_polling_failed;
    case PGRES_POLLING_OK:
      return Qpgres_polling_ok;
    case PGRES_POLLING_READING:
      return Qpgres_polling_reading;
    case PGRES_POLLING_WRITING:
      return Qpgres_polling_writing;
    case PGRES_POLLING_ACTIVE:
      return Qpgres_polling_active;
    default: abort (); /* they've added a new field, die */
    }
}
#endif

DEFUN ("pq-request-cancel", Fpq_request_cancel, 1, 1, 0, /*
Attempt to request cancellation of the current operation.

The return value is t if the cancel request was successfully
dispatched, nil if not (in which case conn->errorMessage is set).
Note: successful dispatch is no guarantee that there will be any effect at
the backend.  The application must read the operation result as usual.
*/
       (conn))
{
  PGconn *P;

  CHECK_PGCONN (conn);
  P = (XPGCONN (conn))->pgconn;

  return PQrequestCancel (P) ? Qt : Qnil;
}


/* accessor function for the PGconn object */
DEFUN ("pq-pgconn", Fpq_pgconn, 2, 2, 0, /*
Accessor function for the PGconn object.
Currently recognized symbols for the field:
pq::db            Database name
pq::user          Database user name
pq::pass          Database user's password
pq::host          Hostname of PostgreSQL backend connected to
pq::port          TCP port number of connection
pq::tty           Debugging TTY (not used in Emacs)
pq::options       Additional backend options
pq::status        Connection status (either OK or BAD)
pq::error-message Last error message from the backend
pq::backend-pid   Process ID of backend process
*/
	(conn, field))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  pgconn = (XPGCONN (conn))->pgconn;

  if (!pgconn)
    error ("Stale connection");
  else if (EQ(field, Qpqdb))
    /* PQdb Returns the database name of the connection.
       char *PQdb(PGconn *conn)
     */
    return build_string (PQdb(pgconn));
  else if (EQ (field, Qpquser))
    /* PQuser Returns the user name of the connection.
       char *PQuser(PGconn *conn)
     */
    return build_string (PQuser(pgconn));
  else if (EQ (field, Qpqpass))
    /* PQpass Returns the password of the connection.
       char *PQpass(PGconn *conn)
     */
    return build_string (PQpass(pgconn));
  else if (EQ (field, Qpqhost))
    /* PQhost Returns the server host name of the connection.
       char *PQhost(PGconn *conn)
     */
    return build_string (PQhost(pgconn));
  else if (EQ (field, Qpqport))
    {
      char *p;
      /* PQport Returns the port of the connection.
         char *PQport(PGconn *conn)
       */
      if ((p = PQport(pgconn)))
	return make_int(atoi(p));
      else
        return make_int(-1);
    }
  else if (EQ (field, Qpqtty))
    /* PQtty Returns the debug tty of the connection.
       char *PQtty(PGconn *conn)
     */
    return build_string (PQtty(pgconn));
  else if (EQ (field, Qpqoptions))
  /* PQoptions Returns the backend options used in the connection.
     char *PQoptions(PGconn *conn)
   */
    return build_string (PQoptions(pgconn));
  else if (EQ (field, Qpqstatus))
    /* PQstatus Returns the status of the connection. The status can be
       CONNECTION_OK or CONNECTION_BAD.
       ConnStatusType PQstatus(PGconn *conn)
     */
    switch (PQstatus (pgconn))
      {
      case CONNECTION_OK: return Qpg_connection_ok;
      case CONNECTION_BAD: return Qpg_connection_bad;
      case CONNECTION_STARTED: return Qpg_connection_started;
      case CONNECTION_MADE: return Qpg_connection_made;
      case CONNECTION_AWAITING_RESPONSE: return Qpg_connection_awaiting_response;
      case CONNECTION_AUTH_OK: return Qpg_connection_auth_ok;
      case CONNECTION_SETENV: return Qpg_connection_setenv;
      default: /* they've added a new field, die */
	abort ();
      }
  else if (EQ (field, Qpqerrormessage))
    /* PQerrorMessage Returns the error message most recently generated
       by an operation on the connection.
       char *PQerrorMessage(PGconn* conn);
     */
    return build_string (PQerrorMessage(pgconn));
  else if (EQ (field, Qpqbackendpid))
    /* PQbackendPID Returns the process ID of the backend server handling
       this connection.
       int PQbackendPID(PGconn *conn);
     */
    return make_int (PQbackendPID(pgconn));
  else
    error ("bad PGconn accessor");
}

/* Query functions */
DEFUN ("pq-exec", Fpq_exec, 2, 2, 0, /*
Submit a query to Postgres and wait for the result.
*/
	(conn, query))
{
  PGconn *pgconn;
  struct Lisp_PGresult *lisp_pgresult;
  PGresult *pgresult;

  CHECK_PGCONN (conn);
  CHECK_STRING (query);

  pgconn = (XPGCONN (conn))->pgconn;

  pgresult = PQexec (pgconn, XSTRING_DATA (query));
  lisp_pgresult = allocate_pgresult();
  lisp_pgresult->pgresult = pgresult;
  lisp_pgresult->live = 1;

  return make_pgresult (lisp_pgresult);
}

DEFUN ("pq-send-query", Fpq_send_query, 2, 2, 0, /*
Submit a query to Postgres and don't wait for the result.
Returns: t if successfully submitted
         nil if error (conn->errorMessage is set)
*/
	(conn, query))
{
  PGconn *pgconn;

  CHECK_PGCONN (conn);
  CHECK_STRING (query);

  pgconn = (XPGCONN (conn))->pgconn;

  return PQsendQuery (pgconn, XSTRING_DATA (query)) ? Qt : Qnil;
}

DEFUN ("pq-get-result", Fpq_get_result, 1, 1, 0, /*
Retrieve an asynchronous result from a query.
NIL is returned when no more query work remains.
*/
	(conn))
{
  PGconn *pgconn;
  struct Lisp_PGresult *lisp_pgresult;
  PGresult *pgresult;

  CHECK_PGCONN (conn);

  pgconn = (XPGCONN (conn))->pgconn;

  pgresult = PQgetResult (pgconn);
  if (!pgresult) return Qnil;

  lisp_pgresult = allocate_pgresult();
  lisp_pgresult->pgresult = pgresult;
  lisp_pgresult->live = 1;

  return make_pgresult (lisp_pgresult);
}

DEFUN ("pq-result-status", Fpq_result_status, 1, 1, 0, /*
Return result status of the query.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  switch (PQresultStatus (pgr)) {
  case PGRES_EMPTY_QUERY: return Qpgres_empty_query;
  case PGRES_COMMAND_OK: return Qpgres_command_ok;
  case PGRES_TUPLES_OK: return Qpgres_tuples_ok;
  case PGRES_COPY_OUT: return Qpgres_copy_out;
  case PGRES_COPY_IN: return Qpgres_copy_in;
  case PGRES_BAD_RESPONSE: return Qpgres_bad_response;
  case PGRES_NONFATAL_ERROR: return Qpgres_nonfatal_error;
  case PGRES_FATAL_ERROR: return Qpgres_fatal_error;
  default: abort ();
  }
}

DEFUN ("pq-res-status", Fpq_res_status, 1, 1, 0, /*
Return stringified result status of the query.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQresStatus(PQresultStatus (pgr)));
}

/* Sundry PGresult accessor functions */
DEFUN ("pq-result-error-message", Fpq_result_error_message, 1, 1, 0, /*
Return last message associated with the query.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQresultErrorMessage(pgr));
}

DEFUN ("pq-ntuples", Fpq_ntuples, 1, 1, 0, /*
Return the number of tuples (instances) in the query result.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQntuples (pgr));
}

DEFUN ("pq-nfields", Fpq_nfields, 1, 1, 0, /*
Return the number of fields (attributes) in each tuple of the query result.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQnfields (pgr));
}

DEFUN ("pq-binary-tuples", Fpq_binary_tuples, 1, 1, 0, /*
Return t if the query result contains binary data, nil otherwise.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return (PQbinaryTuples (pgr)) ? Qt : Qnil;
}

DEFUN ("pq-fname", Fpq_fname, 2, 2, 0, /*
Return the field (attribute) name associated with the given field index.
Field indices start at 0.
*/
	(result, field_index))
{
  /* This function isn't MULE-ized */
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (field_index);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQfname(pgr, XINT (field_index)));
}

DEFUN ("pq-fnumber", Fpq_fnumber, 2, 2, 0, /*
Return the number of fields (attributes) in each tuple of the query result.
*/
	(result, field_name))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_STRING (field_name);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQfnumber (pgr, XSTRING_DATA (field_name)));
}

DEFUN ("pq-ftype", Fpq_ftype, 2, 2, 0, /*
Return the field type associated with the given field index.
The integer returned is the internal coding of the type.  Field indices
start at 0.
*/
	(result, field_num))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (field_num);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQftype (pgr, XINT (field_num)));
}

DEFUN ("pq-fsize", Fpq_fsize, 2, 2, 0, /*
Return the field size in bytes associated with the given field index.
Field indices start at 0.
*/
	(result, field_index))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (field_index);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQftype (pgr, XINT (field_index)));
}

DEFUN ("pq-get-value", Fpq_get_value, 3, 3, 0, /*
Return a single field (attribute) value of one tuple of a PGresult.
Tuple and field indices start at 0.
*/
	(result, tup_num, field_num))
{
  /* This function isn't MULE-ized */
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (tup_num);
  CHECK_INT (field_num);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQgetvalue (pgr, XINT (tup_num), XINT (field_num)));
}

DEFUN ("pq-get-length", Fpq_get_length, 3, 3, 0, /*
Returns the length of a field value in bytes.
If result is binary, i.e. a result of a binary portal, then the
length returned does NOT include the size field of the varlena.  (The
data returned by PQgetvalue doesn't either.)
*/
	(result, tup_num, field_num))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (tup_num);
  CHECK_INT (field_num);
  pgr = (XPGRESULT (result))->pgresult;

  return make_int (PQgetlength (pgr, XINT (tup_num), XINT (field_num)));
}

DEFUN ("pq-get-is-null", Fpq_get_is_null, 3, 3, 0, /*
Returns the null status of a field value.
*/
	(result, tup_num, field_num))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  CHECK_INT (tup_num);
  CHECK_INT (field_num);
  pgr = (XPGRESULT (result))->pgresult;

  return PQgetisnull (pgr, XINT (tup_num), XINT (field_num)) ? Qt : Qnil;
}

DEFUN ("pq-cmd-status", Fpq_cmd_status, 1, 1, 0, /*
Returns the command status string from the SQL command that generated the result.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQcmdStatus (pgr));
}

DEFUN ("pq-cmd-tuples", Fpq_cmd_tuples, 1, 1, 0, /*
Returns the number of rows affected by the SQL command
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

  return build_string (PQcmdTuples (pgr));
}

DEFUN ("pq-oid-value", Fpq_oid_value, 1, 1, 0, /*
Returns the object id of the tuple inserted.
*/
	(result))
{
  PGresult *pgr;

  CHECK_PGRESULT (result);
  pgr = (XPGRESULT (result))->pgresult;

#ifdef HAVE_POSTGRESQLV7
  return make_int (PQoidValue (pgr));
#else
  /* Use the old interface */
  return make_int (atoi (PQoidStatus (pgr)));
#endif
}

#ifdef HAVE_POSTGRESQLV7
DEFUN ("pq-set-nonblocking", Fpq_set_nonblocking, 2, 2, 0, /*
Sets the PGconn's database connection non-blocking if the arg is TRUE
or makes it non-blocking if the arg is FALSE, this will not protect
you from PQexec(), you'll only be safe when using the non-blocking API.

Needs to be called only on a connected database connection.
*/
       (conn, arg))
{
  PGconn *P;

  CHECK_PGCONN (conn);
  P = (XPGCONN (conn))->pgconn;
  return make_int (PQsetnonblocking (P, !NILP (arg)));
}

DEFUN ("pq-is-nonblocking", Fpq_is_nonblocking, 1, 1, 0, /*
Return the blocking status of the database connection
*/
       (conn))
{
  PGconn *P;

  CHECK_PGCONN (conn);
  P = (XPGCONN (conn))->pgconn;
  return PQisnonblocking (P) ? Qt : Qnil;
}

DEFUN ("pq-flush", Fpq_flush, 1, 1, 0, /*
Force the write buffer to be written (or at least try)
*/
       (conn))
{
  PGconn *P;

  CHECK_PGCONN (conn);
  P = (XPGCONN (conn))->pgconn;
  return make_int (PQflush (P));
}
#endif

DEFUN ("pq-notifies", Fpq_notifies, 1, 1, 0, /*
Return the latest async notification that has not yet been handled.
If there has been a notification, then a list of two elements will be returned.
The first element contains the relation name being notified, the second
element contains the backend process ID number.  nil is returned if there
aren't any notifications to process.
*/
       (conn))
{
  /* This function cannot GC */
  PGconn *P;
  PGnotify *PGN;

  CHECK_PGCONN (conn);
  P = (XPGCONN (conn))->pgconn;

  PGN = PQnotifies (P);
  if (!PGN)
    return Qnil;
  else
  {
    Lisp_Object temp;

    temp = list2 (build_string (PGN->relname), make_int (PGN->be_pid));
    free ((void *)PGN);
    return temp;
  }
}

#if defined (HAVE_POSTGRESQLV7) && defined(MULE)
DEFUN ("pq-env-2-encoding", Fpq_env_2_encoding, 0, 0, 0, /*
Get encoding id from environment variable PGCLIENTENCODING.
*/
       ())
{
  return make_int (PQenv2encoding ());
}
#endif /* MULE */

DEFUN ("pq-lo-import", Fpq_lo_import, 2, 2, 0, /*
*/
       (conn, filename))
{
  /* This function hasn't been MULE-ized */
  PGconn *P;

  CHECK_PGCONN (conn);
  CHECK_STRING (filename);

  P = (XPGCONN (conn))->pgconn;

  return (int)lo_import (P, XSTRING_DATA (filename));
}

DEFUN ("pq-lo-export", Fpq_lo_export, 3, 3, 0, /*
*/
       (conn, oid, filename))
{
  /* This function hasn't been MULE-ized */
  PGconn *P;

  CHECK_PGCONN (conn);
  CHECK_INT (oid);
  CHECK_STRING (filename);

  P = (XPGCONN (conn))->pgconn;

  return make_int ((int)lo_export (P, XINT (oid), XSTRING_DATA (filename)));
}

void
syms_of_postgresql(void)
{
  /* opaque exported types */
  defsymbol (&Qpgconnp, "pgconnp");
  defsymbol (&Qpgresultp, "pgresultp");

  /* connection status types */
  defsymbol (&Qpg_connection_ok, "pg::connection-ok");
  defsymbol (&Qpg_connection_bad, "pg::connection-bad");
  defsymbol (&Qpg_connection_started, "pg::connection-started");
  defsymbol (&Qpg_connection_made, "pg::connection-made");
  defsymbol (&Qpg_connection_awaiting_response, "pg::connection-awaiting-response");
  defsymbol (&Qpg_connection_auth_ok, "pg::connection-auth-ok");
  defsymbol (&Qpg_connection_setenv, "pg::connection-setenv");

  /* Fields of PGconn */
  defsymbol (&Qpqdb, "pq::db");
  defsymbol (&Qpquser, "pq::user");
  defsymbol (&Qpqpass, "pq::pass");
  defsymbol (&Qpqhost, "pq::host");
  defsymbol (&Qpqport, "pq::port");
  defsymbol (&Qpqtty, "pq::tty");
  defsymbol (&Qpqoptions, "pq::options");
  defsymbol (&Qpqstatus, "pq::status");
  defsymbol (&Qpqerrormessage, "pq::error-message");
  defsymbol (&Qpqbackendpid, "pq::backend-pid");

  /* Query status results */
  defsymbol (&Qpgres_empty_query, "pgres::empty-query");
  defsymbol (&Qpgres_command_ok, "pgres::command-ok");
  defsymbol (&Qpgres_tuples_ok, "pgres::tuples-ok");
  defsymbol (&Qpgres_copy_out, "pgres::copy-out");
  defsymbol (&Qpgres_copy_in, "pgres::copy-in");
  defsymbol (&Qpgres_bad_response, "pgres::bad-response");
  defsymbol (&Qpgres_nonfatal_error, "pgres::nonfatal-error");
  defsymbol (&Qpgres_fatal_error, "pgres::fatal-error");

  /* Poll status results */
  defsymbol (&Qpgres_polling_failed, "pgres::polling-failed");
  defsymbol (&Qpgres_polling_reading, "pgres::polling-reading");
  defsymbol (&Qpgres_polling_writing, "pgres::polling-writing");
  defsymbol (&Qpgres_polling_ok, "pgres::polling-ok");
  defsymbol (&Qpgres_polling_active, "pgres::polling-active");

#ifdef HAVE_POSTGRESQLV7
  DEFSUBR (Fpq_connect_start);
  DEFSUBR (Fpq_connect_poll);
  DEFSUBR (Fpq_client_encoding);
  DEFSUBR (Fpq_set_client_encoding);
#endif /* HAVE_POSTGRESQLV7 */
  DEFSUBR (Fpq_conn_defaults);
  DEFSUBR (Fpq_connectdb);
  DEFSUBR (Fpq_finish);
  DEFSUBR (Fpq_clear);
  DEFSUBR (Fpq_is_busy);
  DEFSUBR (Fpq_consume_input);

  DEFSUBR (Fpq_reset);
#ifdef HAVE_POSTGRESQLV7
  DEFSUBR (Fpq_reset_start);
  DEFSUBR (Fpq_reset_poll);
#endif
  DEFSUBR (Fpq_request_cancel);
  DEFSUBR (Fpq_pgconn);

  DEFSUBR (Fpq_exec);
  DEFSUBR (Fpq_send_query);
  DEFSUBR (Fpq_get_result);
  DEFSUBR (Fpq_result_status);
  DEFSUBR (Fpq_res_status);
  DEFSUBR (Fpq_result_error_message);
  DEFSUBR (Fpq_ntuples);
  DEFSUBR (Fpq_nfields);
  DEFSUBR (Fpq_binary_tuples);
  DEFSUBR (Fpq_fname);
  DEFSUBR (Fpq_fnumber);
  DEFSUBR (Fpq_ftype);
  DEFSUBR (Fpq_fsize);
  /***/
  DEFSUBR (Fpq_get_value);
  DEFSUBR (Fpq_get_length);
  DEFSUBR (Fpq_get_is_null);
  DEFSUBR (Fpq_cmd_status);
  DEFSUBR (Fpq_cmd_tuples);
  DEFSUBR (Fpq_oid_value);

#ifdef HAVE_POSTGRESQLV7
  DEFSUBR (Fpq_set_nonblocking);
  DEFSUBR (Fpq_is_nonblocking);
  DEFSUBR (Fpq_flush);
#endif
  DEFSUBR (Fpq_notifies);

#if defined (HAVE_POSTGRESQLV7) && defined(MULE)
  DEFSUBR (Fpq_env_2_encoding);
#endif

  DEFSUBR (Fpq_lo_import);
  DEFSUBR (Fpq_lo_export);
}

void
vars_of_postgresql(void)
{
  char *p;

  Fprovide (intern ("postgresql"));
#ifdef HAVE_POSTGRESQLV7
  Fprovide (intern ("postgresqlv7"));
#endif

  if ((p = getenv("PGHOST")))
    {
      VXPGHOST = build_string(p);
    }
  else
    {
      VXPGHOST = Qnil;
    }
  DEFVAR_LISP ("pg:host", &VXPGHOST /*
Default PostgreSQL server name.
If not set, the server running on the local host is used.  The
initial value is set from the PGHOST environment variable.
*/ );

  if ((p = getenv("PGUSER")))
    {
      VXPGUSER = build_string(p);
    }
  else
    {
      VXPGUSER = Qnil;
    }
  DEFVAR_LISP ("pg:user", &VXPGUSER /*
Default PostgreSQL user name.
This value is used when connecting to a database for authentication.
The initial value is set from the PGUSER environment variable.
*/ );

  if ((p = getenv("PGOPTIONS")))
    {
      VXPGOPTIONS = build_string(p);
    }
  else
    {
      VXPGOPTIONS = Qnil;
    }
  DEFVAR_LISP ("pg:options", &VXPGOPTIONS /*
Default PostgreSQL user name.
This value is used when connecting to a database for authentication.
The initial value is set from the PGUSER environment variable.
*/ );

  if ((p = getenv("PGPORT")))
    {
      VXPGPORT = make_int(atoi(p));
    }
  else
    {
      VXPGPORT = Qnil;
    }
  DEFVAR_LISP ("pg:port", &VXPGPORT /*
Default port to connect to PostgreSQL backend.
This value is used when connecting to a database.
The initial value is set from the PGPORT environment variable.
*/ );

  if ((p = getenv("PGTTY")))
    {
      VXPGTTY = build_string (p);
    }
  else
    {
      VXPGTTY = Qnil;
    }
  DEFVAR_LISP ("pg:tty", &VXPGTTY /*
Default debugging TTY.
There is no useful setting of this variable in the XEmacs Lisp API.
The initial value is set from the PGTTY environment variable.
*/ );

  if ((p = getenv("PGDATABASE")))
    {
      VXPGDATABASE = build_string(p);
    }
  else
    {
      VXPGDATABASE = Qnil;
    }
  DEFVAR_LISP ("pg:database", &VXPGDATABASE /*
Default database to connect to.
The initial value is set from the PGDATABASE environment variable.
*/ );

  if ((p = getenv("PGREALM")))
    {
      VXPGREALM = build_string(p);
    }
  else
    {
      VXPGREALM = Qnil;
    }
  DEFVAR_LISP ("pg:realm", &VXPGREALM /*
Default kerberos realm to use for authentication.
The initial value is set from the PGREALM environment variable.
*/ );

#ifdef MULE
  if ((p = getenv("PGCLIENTENCODING")))
    {
      VXPGCLIENTENCODING = build_string(p);
    }
  else
    {
      VXPGCLIENTENCODING = Qnil;
    }
  DEFVAR_LISP ("pg:client-encoding", &VXPGCLIENTENCODING /*
Default client encoding to use.
The initial value is set from the PGCLIENTENCODING environment variable.
*/ );
#endif

#if !defined(HAVE_POSTGRESQLV7)
  if ((p = getenv("PGAUTHTYPE")))
    {
      VXPGAUTHTYPE = build_string(p);
    }
  else
    {
      VXPGAUTHTYPE = Qnil;
    }
  DEFVAR_LISP ("pg:authtype", &VXPGAUTHTYPE /*
Default authentication to use.
The initial value is set from the PGAUTHTYPE environment variable.

WARNING:  This variable has gone away in versions of PostgreSQL newer
than 6.5.
*/ );
#endif

  if ((p = getenv("PGGEQO")))
    {
      VXPGGEQO = build_string(p);
    }
  else
    {
      VXPGGEQO = Qnil;
    }
  DEFVAR_LISP ("pg:geqo", &VXPGGEQO /*
Genetic Query Optimizer options.
The initial value is set from the PGGEQO environment variable.
*/ );

  if ((p = getenv("PGCOSTINDEX")))
    {
      VXPGCOSTINDEX = build_string(p);
    }
  else
    {
      VXPGCOSTINDEX = Qnil;
    }
  DEFVAR_LISP ("pg:cost-index", &VXPGCOSTINDEX /*
Default cost index options.
The initial value is set from the PGCOSTINDEX environment variable.
*/ );

  if ((p = getenv("PGCOSTHEAP")))
    {
      VXPGCOSTHEAP = build_string(p);
    }
  else
    {
      VXPGCOSTHEAP = Qnil;
    }
  DEFVAR_LISP ("pg:cost-heap", &VXPGCOSTHEAP /*
Default cost heap options.
The initial value is set from the PGCOSTHEAP environment variable.
*/ );

  if ((p = getenv("PGTZ")))
    {
      VXPGTZ = build_string(p);
    }
  else
    {
      VXPGTZ = Qnil;
    }
  DEFVAR_LISP ("pg:tz", &VXPGTZ /*
Default timezone to use.
The initial value is set from the PGTZ environment variable.
*/ );

  if ((p = getenv("PGDATESTYLE")))
    {
      VXPGDATESTYLE = build_string(p);
    }
  else
    {
      VXPGDATESTYLE = Qnil;
    }
  DEFVAR_LISP ("pg:date-style", &VXPGDATESTYLE /*
Default date style to use.
The initial value is set from the PGDATESTYLE environment variable.
*/ );
}
